import { XMLDate, XMLElement, XMLBoolean, XMLNumber, XMLList, INamespaced, ISchemaInformation } from "./BaseTypes";
import { ensureArray } from "./Util";

const schemaVersion = "2.1.0";
const namespace = "si";
const namespaceUrl = `https://ptb.de/${namespace}`;
const schemaLocation = `${namespaceUrl}/v${schemaVersion}/SI_Format.xsd`;
export const DSI: ISchemaInformation = {
  schemaVersion,
  namespace,
  namespaceUrl,
  schemaLocation,
};

export class SIXMLElement extends XMLElement implements INamespaced {
  get namespace() {
    return DSI.namespace;
  }
}

export class SIXMLBoolean extends XMLBoolean implements INamespaced {
  get namespace() {
    return DSI.namespace;
  }
}

export class SIXMLDate extends XMLDate implements INamespaced {
  get namespace() {
    return DSI.namespace;
  }
}

export class SIXMLNumber extends XMLNumber implements INamespaced {
  get namespace() {
    return DSI.namespace;
  }
}

export class SIXMLList extends XMLList implements INamespaced {
  get namespace() {
    return DSI.namespace;
  }
}

export class RealQuantityType extends SIXMLElement implements INamespaced {
  label?: SIXMLElement;
  value: SIXMLNumber;
  unit: SIXMLElement;
  dateTime?: SIXMLDate;
  expandedUnc?: ExpandedUncType;
  coverageInterval?: CoverageIntervalType;

  constructor(el: Partial<RealQuantityType> = {}) {
    super(el);
    if (el.label) this.label = new SIXMLElement(el.label);
    this.value = new SIXMLNumber(el.value);
    this.unit = new SIXMLElement(el.unit);
    if (el.dateTime) this.dateTime = new SIXMLDate(el.dateTime);
    if (el.expandedUnc) this.expandedUnc = new ExpandedUncType(el.expandedUnc);
    if (el.coverageInterval) this.coverageInterval = new CoverageIntervalType(el.coverageInterval);
  }
}

export class HybridType extends SIXMLElement {
  real: RealQuantityType[];
  complex: ComplexQuantityType[];
  list: SIListType[];
  realList: RealListType[];
  realListXMLList: RealListXMLListType[];
  complexList: ComplexListType[];
  constant: ConstantQuantityType[];

  constructor(el: Partial<HybridType> = {}) {
    super(el);
    this.real = ensureArray(el.real).map((x) => new RealQuantityType(x));
    this.complex = ensureArray(el.complex).map((x) => new ComplexQuantityType(x));
    this.list = ensureArray(el.list).map((x) => new SIListType(x));
    this.realList = ensureArray(el.realList).map((x) => new RealListType(x));
    this.realListXMLList = ensureArray(el.realListXMLList).map((x) => new RealListXMLListType(x));
    this.complexList = ensureArray(el.complexList).map((x) => new ComplexListType(x));
    this.constant = ensureArray(el.constant).map((x) => new ConstantQuantityType(x));
  }
}

export class ComplexQuantityType extends SIXMLElement {
  label?: SIXMLElement;
  dateTime?: SIXMLDate;
  valueReal: SIXMLNumber;
  valueImag: SIXMLNumber;
  unit: SIXMLElement[];
  valueMagnitude: SIXMLNumber;
  valuePhase: SIXMLNumber;
  unitPhase: SIXMLElement;
  ellipsoidalRegion?: EllipsoidalRegionType;
  rectangularRegion?: RectangularRegionType;

  constructor(el: Partial<ComplexQuantityType> = {}) {
    super(el);
    if (el.label) this.label = new SIXMLElement(el.label);
    if (el.dateTime) this.dateTime = new SIXMLDate(el.dateTime);
    this.valueReal = new SIXMLNumber(el.valueReal);
    this.valueImag = new SIXMLNumber(el.valueImag);
    this.unit = ensureArray(el.unit).map((x) => new SIXMLElement(x));
    this.valueMagnitude = new SIXMLNumber(el.valueMagnitude);
    this.valuePhase = new SIXMLNumber(el.valuePhase);
    this.unitPhase = new SIXMLElement(el.unitPhase);
    if (el.ellipsoidalRegion) this.ellipsoidalRegion = new EllipsoidalRegionType(el.ellipsoidalRegion);
    if (el.rectangularRegion) this.rectangularRegion = new RectangularRegionType(el.rectangularRegion);
  }
}

export class ConstantQuantityType extends SIXMLElement {
  label?: SIXMLElement;
  value: SIXMLNumber;
  unit: SIXMLElement;
  dateTime?: SIXMLDate;
  uncertainty?: SIXMLNumber;
  distribution?: SIXMLElement;

  constructor(el: Partial<ConstantQuantityType> = {}) {
    super(el);
    if (el.label) this.label = new SIXMLElement(el.label);
    this.value = new SIXMLNumber(el.value);
    this.unit = new SIXMLElement(el.unit);
    if (el.dateTime) this.dateTime = new SIXMLDate(el.dateTime);
    if (el.uncertainty) this.uncertainty = new SIXMLNumber(el.uncertainty);
    if (el.distribution) this.distribution = new SIXMLElement(el.distribution);
  }
}

export class RealListXMLListType extends SIXMLElement {
  labelXMLList?: StringXMLListType;
  valueXMLList: DecimalXMLListType;
  unitXMLList: UnitXMLListType;
  dateTimeXMLList?: DateTimeXMLListType;
  expandedUncXMLList?: ExpandedUncXMLListType;
  coverageIntervalXMLList?: CoverageIntervalXMLListType;
  ellipsoidalRegionXMLList?: EllipsoidalRegionType;
  rectangularRegionXMLList?: RectangularRegionType;

  constructor(el: Partial<RealListXMLListType> = {}) {
    super(el);
    if (el.labelXMLList) this.labelXMLList = new StringXMLListType(el.labelXMLList);
    this.valueXMLList = new DecimalXMLListType(el.valueXMLList);
    this.unitXMLList = new UnitXMLListType(el.unitXMLList);
    if (el.dateTimeXMLList) this.dateTimeXMLList = new DateTimeXMLListType(el.dateTimeXMLList);
    if (el.expandedUncXMLList) this.expandedUncXMLList = new ExpandedUncXMLListType(el.expandedUncXMLList);
    if (el.coverageIntervalXMLList) this.coverageIntervalXMLList = new CoverageIntervalXMLListType(el.coverageIntervalXMLList);
    if (el.ellipsoidalRegionXMLList) this.ellipsoidalRegionXMLList = new EllipsoidalRegionType(el.ellipsoidalRegionXMLList);
    if (el.rectangularRegionXMLList) this.rectangularRegionXMLList = new RectangularRegionType(el.rectangularRegionXMLList);
  }
}

export class SIListType extends SIXMLElement {
  label?: SIXMLElement;
  dateTime?: SIXMLDate;
  realList: RealListType[];
  complexList: ComplexListType[];
  list: SIListType[];

  constructor(el: Partial<SIListType> = {}) {
    super(el);
    if (el.label) this.label = new SIXMLElement(el.label);
    if (el.dateTime) this.dateTime = new SIXMLDate(el.dateTime);
    this.realList = ensureArray(el.realList).map((x) => new RealListType(x));
    this.complexList = ensureArray(el.complexList).map((x) => new ComplexListType(x));
    this.list = ensureArray(el.list).map((x) => new SIListType(x));
  }
}

export class ExpandedUncType extends SIXMLElement {
  uncertainty: SIXMLNumber;
  coverageFactor: SIXMLNumber;
  coverageProbability: SIXMLNumber;
  distribution?: SIXMLElement;

  constructor(el: Partial<ExpandedUncType> = {}) {
    super(el);
    this.uncertainty = new SIXMLNumber(el.uncertainty);
    this.coverageFactor = new SIXMLNumber(el.coverageFactor);
    this.coverageProbability = new SIXMLNumber(el.coverageProbability);
    if (el.distribution) this.distribution = new SIXMLElement(el.distribution);
  }
}

export class CoverageIntervalType extends SIXMLElement {
  standardUnc: SIXMLNumber;
  intervalMin: SIXMLNumber;
  intervalMax: SIXMLNumber;
  coverageProbability: SIXMLNumber;
  distribution?: SIXMLElement;

  constructor(el: Partial<CoverageIntervalType> = {}) {
    super(el);
    this.standardUnc = new SIXMLNumber(el.standardUnc);
    this.intervalMin = new SIXMLNumber(el.intervalMin);
    this.intervalMax = new SIXMLNumber(el.intervalMax);
    this.coverageProbability = new SIXMLNumber(el.coverageProbability);
    if (el.distribution) this.distribution = new SIXMLElement(el.distribution);
  }
}

export class EllipsoidalRegionType extends SIXMLElement {
  covarianceMatrix: CovarianceMatrixType;
  coverageFactor: SIXMLNumber;
  coverageProbability: SIXMLNumber;
  distribution?: SIXMLElement;

  constructor(el: Partial<EllipsoidalRegionType> = {}) {
    super(el);
    this.covarianceMatrix = new CovarianceMatrixType(el.covarianceMatrix);
    this.coverageFactor = new SIXMLNumber(el.coverageFactor);
    this.coverageProbability = new SIXMLNumber(el.coverageProbability);
    if (el.distribution) this.distribution = new SIXMLElement(el.distribution);
  }
}

export class RectangularRegionType extends SIXMLElement {
  covarianceMatrix: CovarianceMatrixType;
  coverageFactor: SIXMLNumber;
  coverageProbability: SIXMLNumber;
  distribution?: SIXMLElement;

  constructor(el: Partial<RectangularRegionType> = {}) {
    super(el);
    this.covarianceMatrix = new CovarianceMatrixType(el.covarianceMatrix);
    this.coverageFactor = new SIXMLNumber(el.coverageFactor);
    this.coverageProbability = new SIXMLNumber(el.coverageProbability);
    if (el.distribution) this.distribution = new SIXMLElement(el.distribution);
  }
}

export class RealInListType extends SIXMLElement {
  label?: SIXMLElement;
  value: SIXMLNumber;
  unit?: SIXMLElement;
  dateTime?: SIXMLDate;
  expandedUnc?: ExpandedUncType;
  coverageInterval?: CoverageIntervalType;

  constructor(el: Partial<RealInListType> = {}) {
    super(el);
    if (el.label) this.label = new SIXMLElement(el.label);
    this.value = new SIXMLNumber(el.value);
    if (el.unit) this.unit = new SIXMLElement(el.unit);
    if (el.dateTime) this.dateTime = new SIXMLDate(el.dateTime);
    if (el.expandedUnc) this.expandedUnc = new ExpandedUncType(el.expandedUnc);
    if (el.coverageInterval) this.coverageInterval = new CoverageIntervalType(el.coverageInterval);
  }
}

export class ComplexInListType extends SIXMLElement {
  label?: SIXMLElement;
  dateTime?: SIXMLDate;
  valueReal: SIXMLNumber;
  valueImag: SIXMLNumber;
  unit?: SIXMLElement[];
  valueMagnitude: SIXMLNumber;
  valuePhase: SIXMLNumber;
  unitPhase?: SIXMLElement;
  ellipsoidalRegion?: EllipsoidalRegionType;
  rectangularRegion?: RectangularRegionType;

  constructor(el: Partial<ComplexInListType> = {}) {
    super(el);
    if (el.label) this.label = new SIXMLElement(el.label);
    if (el.dateTime) this.dateTime = new SIXMLDate(el.dateTime);
    this.valueReal = new SIXMLNumber(el.valueReal);
    this.valueImag = new SIXMLNumber(el.valueImag);
    if (el.unit) this.unit = ensureArray(el.unit).map((x) => new SIXMLElement(x));
    this.valueMagnitude = new SIXMLNumber(el.valueMagnitude);
    this.valuePhase = new SIXMLNumber(el.valuePhase);
    if (el.unitPhase) this.unitPhase = new SIXMLElement(el.unitPhase);
    if (el.ellipsoidalRegion) this.ellipsoidalRegion = new EllipsoidalRegionType(el.ellipsoidalRegion);
    if (el.rectangularRegion) this.rectangularRegion = new RectangularRegionType(el.rectangularRegion);
  }
}

export class RealListType extends SIXMLElement {
  label?: SIXMLElement;
  dateTime?: SIXMLDate;
  listUnit?: SIXMLElement;
  /** Definition of a structure, for a global univariate uncertainty, that
   * is used within the list structure with a list of real quantities.
   * The global univariate uncertainty can either be given as an expanded
   * measurement uncertainty or as a coverage interval. */
  listUnivariateUnc?: ListUnivariateUncType;
  /** Meta data element definition for a real measurement quantity in list.
   * This implementation differs from the pure real quantity in the way that
   * the unit component is optional in order to allow a combination with
   * a global unit in the list of real quantities. */
  real: RealInListType[];
  ellipsoidalRegion?: EllipsoidalRegionType;
  rectangularRegion?: RectangularRegionType;

  constructor(el: Partial<RealListType> = {}) {
    super(el);
    if (el.label) this.label = new SIXMLElement(el.label);
    if (el.dateTime) this.dateTime = new SIXMLDate(el.dateTime);
    if (el.listUnit) this.listUnit = new SIXMLElement(el.listUnit);
    if (el.listUnivariateUnc) this.listUnivariateUnc = new ListUnivariateUncType(el.listUnivariateUnc);
    this.real = ensureArray(el.real).map((x) => new RealInListType(x));
    if (el.ellipsoidalRegion) this.ellipsoidalRegion = new EllipsoidalRegionType(el.ellipsoidalRegion);
    if (el.rectangularRegion) this.rectangularRegion = new RectangularRegionType(el.rectangularRegion);
  }
}

export class ListUnivariateUncType extends SIXMLElement {
  expandedUnc: ExpandedUncType;
  coverageInterval: CoverageIntervalType;

  constructor(el: Partial<ListUnivariateUncType> = {}) {
    super(el);
    this.expandedUnc = new ExpandedUncType(el.expandedUnc);
    this.coverageInterval = new CoverageIntervalType(el.coverageInterval);
  }
}

/** Type providing listing of strings (xs:string)
 * with separation by blank spaces. */
export class StringXMLListType extends SIXMLList {}

/** Type providing listing of values (si:decimalType)
 * with separation by blank spaces. */
export class DecimalXMLListType extends SIXMLList {}

/** Type providing listing of units (si:unitType)
 * with separation by blank spaces. */
export class UnitXMLListType extends SIXMLList {}

/** Type providing listing of time values (xs:dateTime)
 * with separation by blank spaces. */
export class DateTimeXMLListType extends SIXMLList {}

export class ExpandedUncXMLListType extends SIXMLElement {
  uncertaintyXMLList: UncertaintyValueXMLListType;
  coverageFactorXMLList: KValueXMLListType;
  coverageProbabilityXMLList: ProbabilityValueXMLListType;
  distributionXMLList?: StringXMLListType;

  constructor(el: Partial<ExpandedUncXMLListType> = {}) {
    super(el);
    this.uncertaintyXMLList = new UncertaintyValueXMLListType(el.uncertaintyXMLList);
    this.coverageFactorXMLList = new KValueXMLListType(el.coverageFactorXMLList);
    this.coverageProbabilityXMLList = new ProbabilityValueXMLListType(el.coverageProbabilityXMLList);
    if (el.distributionXMLList) this.distributionXMLList = new StringXMLListType(el.distributionXMLList);
  }
}

export class CoverageIntervalXMLListType extends SIXMLElement {
  standardUncXMLList: UncertaintyValueXMLListType;
  intervalMinXMLList: DecimalXMLListType;
  intervalMaxXMLList: DecimalXMLListType;
  coverageProbabilityXMLList: ProbabilityValueXMLListType;
  distributionXMLList?: StringXMLListType;

  constructor(el: Partial<CoverageIntervalXMLListType> = {}) {
    super(el);
    this.standardUncXMLList = new UncertaintyValueXMLListType(el.standardUncXMLList);
    this.intervalMinXMLList = new DecimalXMLListType(el.intervalMinXMLList);
    this.intervalMaxXMLList = new DecimalXMLListType(el.intervalMaxXMLList);
    this.coverageProbabilityXMLList = new ProbabilityValueXMLListType(el.coverageProbabilityXMLList);
    if (el.distributionXMLList) this.distributionXMLList = new StringXMLListType(el.distributionXMLList);
  }
}

export class ComplexListType extends SIXMLElement {
  label?: SIXMLElement;
  dateTime?: SIXMLDate;
  listUnit?: SIXMLElement;
  listUnitPhase?: SIXMLElement;
  /** Definition of a structure, for a global bivariate uncertainty, that
   * is used within the list structure with a list of complex quantities.
   * The global bivariate uncertainty can either be given as a hyper-ellipsoidal
   * coverage region or a hyper-rectangular coverage region. Both
   * coverage regions must provide a covariance matrix of dimension 2. */
  listBivariateUnc?: ListBivariateUncType;
  /** Meta data element definition for a complex measurement quantity in list.
   * This implementation differs from the pure complex quantity, in the way that
   * the unit components are optional in order to allow a combination with
   * a global unit in the list of complex quantities. */
  complex: ComplexInListType[];
  ellipsoidalRegion?: EllipsoidalRegionType;
  rectangularRegion?: RectangularRegionType;

  constructor(el: Partial<ComplexListType> = {}) {
    super(el);
    if (el.label) this.label = new SIXMLElement(el.label);
    if (el.dateTime) this.dateTime = new SIXMLDate(el.dateTime);
    if (el.listUnit) this.listUnit = new SIXMLElement(el.listUnit);
    if (el.listUnitPhase) this.listUnitPhase = new SIXMLElement(el.listUnitPhase);
    if (el.listBivariateUnc) this.listBivariateUnc = new ListBivariateUncType(el.listBivariateUnc);
    this.complex = ensureArray(el.complex).map((x) => new ComplexInListType(x));
    if (el.ellipsoidalRegion) this.ellipsoidalRegion = new EllipsoidalRegionType(el.ellipsoidalRegion);
    if (el.rectangularRegion) this.rectangularRegion = new RectangularRegionType(el.rectangularRegion);
  }
}

export class ListBivariateUncType extends SIXMLElement {
  ellipsoidalRegion?: EllipsoidalRegionType;
  rectangularRegion?: RectangularRegionType;

  constructor(el: Partial<ListBivariateUncType> = {}) {
    super(el);
    if (el.ellipsoidalRegion) this.ellipsoidalRegion = new EllipsoidalRegionType(el.ellipsoidalRegion);
    if (el.rectangularRegion) this.rectangularRegion = new RectangularRegionType(el.rectangularRegion);
  }
}

/** Type providing listing of uncertainty values (si:uncertaintyValueType)
 * with separation by blank spaces. */
export class UncertaintyValueXMLListType extends SIXMLList {}

/** Type providing listing of coverage factor values (si:kValueType)
 * with separation by blank spaces. */
export class KValueXMLListType extends SIXMLList {}

/** Type providing listing of coverage probability values (si:probabilityValueType)
 * with separation by blank spaces. */
export class ProbabilityValueXMLListType extends SIXMLList {}

export class CovarianceMatrixType extends SIXMLElement {
  /** Definition of a column in the covariance matrix.
   *
   * The column has n covariance elements, where
   * n is the amount of columns in the covariance matrix.
   *
   * The covariance elements are ordered as in the covariance matrix, started
   * at the element in the first row of the matrix and going to the
   * last row of the matrix. */
  column: CovarianceMatrixTypeColumnType[];

  constructor(el: Partial<CovarianceMatrixType> = {}) {
    super(el);
    this.column = ensureArray(el.column).map((x) => new CovarianceMatrixTypeColumnType(x));
  }
}

export class CovarianceMatrixTypeColumnType extends SIXMLElement {
  /** Each covariance component is defined by
   * - element value    (decimal value type)
   * - element unit    (string - SI format) */
  covariance: CovarianceMatrixTypeColumnTypeCovarianceType[];

  constructor(el: Partial<CovarianceMatrixTypeColumnType> = {}) {
    super(el);
    this.covariance = ensureArray(el.covariance).map((x) => new CovarianceMatrixTypeColumnTypeCovarianceType(x));
  }
}

export class CovarianceMatrixTypeColumnTypeCovarianceType extends SIXMLElement {
  value: SIXMLNumber;
  unit: SIXMLElement;

  constructor(el: Partial<CovarianceMatrixTypeColumnTypeCovarianceType> = {}) {
    super(el);
    this.value = new SIXMLNumber(el.value);
    this.unit = new SIXMLElement(el.unit);
  }
}
